/*
** pork_buddy_list.c - Buddy list screen widget.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_status.h>
#include <pork_opt.h>
#include <pork_format.h>
#include <pork_color.h>

extern struct screen screen;

static void blist_cell_destroy_cb(void *d) {
	struct slist_cell *cell = d;

	if (cell->type == TYPE_LIST_CELL) {
		struct bgroup *group = cell->data;

		group->blist_line = NULL;
	} else {
		struct buddy *buddy= cell->data;

		buddy->blist_line = NULL;
	}
}

static void blist_make_indicator(struct buddy *buddy, char *buf, size_t len) {
	switch (buddy->status) {
		case STATUS_AWAY:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_AWAY), len);
			break;

		case STATUS_IDLE:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_IDLE), len);
			break;

		case STATUS_ACTIVE:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_ACTIVE), len);
			break;

		default:
			xstrncpy(buf, "%p?%x", len);
			break;
	}
}

static void blist_make_idle(struct buddy *buddy, char *buf, size_t len) {
	u_int32_t i;
	u_int32_t j;
	u_int32_t s = 0;
	char *format = opt_get_str(OPT_FORMAT_BLIST_IDLE);
	char tmp[32];

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			if (format[++i] == 'I') {
				time_to_str(buddy->idle_time, tmp, sizeof(tmp));
				for (j = 0 ; tmp[j] != '\0' && s < len && j < sizeof(tmp) ; j++)
					buf[s++] = tmp[j];
			} else {
				buf[s++] = format[--i];
			}
		} else {
			buf[s++] = format[i];
		}
	}
	buf[s] = '\0';
}

static void blist_make_warn(struct buddy *buddy, char *buf, size_t len) {
	u_int32_t i;
	u_int32_t j;
	u_int32_t s = 0;
	char *format = opt_get_str(OPT_FORMAT_BLIST_WARN);
	char tmp[32];

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			if (format[++i] == 'W') {
				snprintf(tmp, sizeof(tmp), "%u", buddy->warn_level);
				for (j = 0 ; tmp[j] != '\0' && s < len && j < sizeof(tmp) ; j++)
					buf[s++] = tmp[j];
			} else {
				buf[s++] = format[--i];
			}
		} else {
			buf[s++] = format[i];
		}
	}
	buf[s] = '\0';
}

static void blist_make_buddy_label(	struct blist *blist __notused,
									struct buddy *buddy,
									chtype *ch,
									size_t len)
{
	u_int32_t i;
	u_int32_t s = 0;
	char tmp[32];
	char *format = opt_get_str(OPT_FORMAT_BLIST);
	char *label_left = xmalloc(len);
	char *label_right = NULL;
	char *label = label_left;

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			*tmp = '\0';
			switch (format[++i]) {
				case 'B':
				case 'b':
					blist_make_indicator(buddy, tmp, sizeof(tmp));
					break;

				case 'N':
				case 'n':
					xstrncpy(tmp, buddy->name, sizeof(tmp));
					break;

				case 'I':
					blist_make_idle(buddy, tmp, sizeof(tmp));
					break;

				case 'i':
					if (buddy->idle_time > 0)
						blist_make_idle(buddy, tmp, sizeof(tmp));
					break;

				case 'W':
					blist_make_warn(buddy, tmp, sizeof(tmp));
					break;

				case 'w':
					if (buddy->warn_level > 0)
						blist_make_warn(buddy, tmp, sizeof(tmp));
					break;

				case '>':
					if (label_right == NULL) {
						label[s] = '\0';
						label_right = xmalloc(len + 1);
						label = label_right;
						s = 0;
						break;
					}

				default:
					label[s++] = format[--i];
					break;

			}

			if (*tmp != '\0') {
				u_int32_t j = 0;

				while (tmp[j] != '\0' && j < sizeof(tmp) && s < len)
					label[s++] = tmp[j++];
			}

		} else {
			label[s++] = format[i];
		}
	}

	label[s] = '\0';

	if (label_right == NULL) {
		plaintext_to_cstr(label, ch, len);
		free(label);
	} else {
		size_t len_l;
		size_t len_r;
		chtype *chl;
		chtype *chr;
		u_int32_t i = 0;
		u_int32_t j = 0;
		u_int32_t fill_len;

		len_l = strlen(label_left) + 1;
		len_r = strlen(label_right) + 1;

		chl = xmalloc(sizeof(chtype) * len_l);
		chr = xmalloc(sizeof(chtype) * len_r);

		plaintext_to_cstr(label_left, chl, len_l);
		plaintext_to_cstr(label_right, chr, len_r);

		while (i < len && chl[i] != 0) {
			ch[i] = chl[i];
			i++;
		}

		fill_len = blist->slist.cols - i - cstrlen(chr) - 2;
		while (i < len && fill_len > 0) {
			ch[i++] = ' ';
			fill_len--;
		}

		while (i < len && chr[j] != 0)
			ch[i++] = chr[j++];

		ch[i] = 0;

		free(chl);
		free(chr);
		free(label_left);
		free(label_right);
	}
}

static void blist_make_group_label(	struct blist *blist __notused,
									struct bgroup *group,
									chtype *ch,
									size_t len)
{
	char *label = xmalloc(len);
	char expand_char = '-';
	dlist_t *node = group->blist_line;

	if (node != NULL) {
		struct slist_cell *cell = node->data;

		if (cell != NULL && cell->collapsed)
			expand_char = '+';
	}

	snprintf(label, len, "%c %s (%u/%u)",
		expand_char, group->name, group->num_online,
		group->num_members);

	plaintext_to_cstr(label, ch, len);
	free(label);
}

void blist_update_label(struct blist *blist, dlist_t *node) {
	struct slist_cell *cell;

	if (blist == NULL || node == NULL)
		return;
	cell = node->data;

	if (cell->type == TYPE_FLAT_CELL) {
		struct buddy *buddy = cell->data;

		blist_make_buddy_label(blist, buddy, cell->label, blist->label_len);
	} else {
		struct bgroup *group = cell->data;

		blist_make_group_label(blist, group, cell->label, blist->label_len);
	}

	if (cell->line >= 0) {
		if (node == blist->slist.cursor)
			blist_draw_cursor(blist, 1);
		else
			blist_draw_line(blist, cell);
	}
}

int blist_init(struct aim_priv *priv) {
	int cols = DEFAULT_BLIST_WIDTH;
	struct blist *blist;
	dlist_t *cur;
	int ret;

	if (priv == NULL || priv->blist != NULL)
		return (-1);

	blist = xcalloc(1, sizeof(*blist));

	ret = slist_init(&blist->slist, screen.rows - STATUS_ROWS, cols,
			screen.cols - cols, 0);
	if (ret != 0) {
		free(blist);
		return (-1);
	}

	slist_cell_free_cb(&blist->slist, blist_cell_destroy_cb);

	blist->label_len = SLIST_LAST_COL(&blist->slist) + 1;

	/* Add the buddy groups to the buddy list */
	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct bgroup *group = cur->data;
		dlist_t *node = blist_add_group(blist, group);

		if (node != NULL)
			group->blist_line = node;

		cur = cur->next;
	}

	blist_draw(blist);
	priv->blist = blist;
	return (0);
}

dlist_t *blist_add_group(struct blist *blist, struct bgroup *group) {
	struct slist_cell *cell;
	dlist_t *ret;

	if (blist == NULL)
		return (NULL);

	cell = xcalloc(1, sizeof(*cell));
	cell->type = TYPE_LIST_CELL;
	cell->parent = NULL;
	cell->children = NULL;
	cell->collapsed = 0;
	cell->refnum = group->refnum;
	cell->label = xmalloc(sizeof(chtype) * blist->label_len);
	cell->data = group;
	blist_make_group_label(blist, group, cell->label, blist->label_len);

	ret = slist_add(&blist->slist, cell);

	if (ret == NULL)
		return (NULL);

	group->blist_line = ret;
	blist_draw(blist);
	return (ret);
}

dlist_t *blist_add(struct blist *blist, struct buddy *buddy) {
	struct slist_cell *cell;
	dlist_t *ret;
	dlist_t *node;

	if (blist == NULL)
		return (NULL);

	cell = xcalloc(1, sizeof(*cell));
	cell->type = TYPE_FLAT_CELL;
	cell->parent = buddy->group->blist_line;
	cell->refnum = buddy->refnum;
	cell->label = xmalloc(sizeof(chtype) * blist->label_len);
	cell->data = buddy;
	blist_make_buddy_label(blist, buddy, cell->label, blist->label_len);

	ret = slist_add(&blist->slist, cell);
	if (ret == NULL)
		return (NULL);

	node = buddy->group->blist_line;
	cell = node->data;

	buddy->blist_line = ret;
	blist_make_group_label(blist, buddy->group, cell->label, blist->label_len);
	blist_draw(blist);
	return (ret);
}

void blist_del(struct blist *blist, struct buddy *buddy) {
	dlist_t *node;
	struct slist_cell *cell;

	if (blist == NULL)
		return;

	node = buddy->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;
	buddy->blist_line = NULL;
	slist_del(&blist->slist, cell);

	node = buddy->group->blist_line;
	cell = node->data;
	blist_make_group_label(blist, buddy->group, cell->label, blist->label_len);

	blist_draw(blist);
}

void blist_del_group(struct blist *blist, struct bgroup *group) {
	dlist_t *node;
	struct slist_cell *cell;

	if (blist == NULL)
		return;

	node = group->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;
	group->blist_line = NULL;

	slist_del(&blist->slist, cell);
	blist_draw(blist);
}

void blist_collapse_group(struct blist *blist, struct bgroup *group) {
	dlist_t *node;
	struct slist_cell *cell;

	node = group->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;

	if (!cell->collapsed) {
		if (slist_collapse_list_cell(&blist->slist, group->blist_line) != 0)
			return;
	} else {
		dlist_t *cur = group->member_list;

		cell->collapsed = 0;

		while (cur != NULL) {
			struct buddy *buddy = cur->data;

			if (buddy->status != STATUS_OFFLINE)
				blist_add(blist, buddy);

			cur = cur->next;
		}
	}

	blist_make_group_label(blist, group, cell->label, blist->label_len);
	blist_draw(blist);
}

int blist_cursor_up(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_up(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

int blist_cursor_down(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_down(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

int blist_cursor_start(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_start(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

int blist_cursor_end(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_end(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

int blist_cursor_pgdown(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_pgdown(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

int blist_cursor_pgup(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_pgup(&blist->slist);
	blist_draw_cursor(blist, 1);

	return (ret);
}

inline struct slist_cell *blist_get_cursor(struct blist *blist) {
	if (blist == NULL)
		return (NULL);

	return (slist_get_cursor(&blist->slist));
}

int blist_resize(	struct blist *blist,
					u_int32_t rows,
					u_int32_t cols,
					u_int32_t screen_cols)
{
	int ret;
	int needs_update = 0;

	if (blist == NULL)
		return (0);

	if (blist->slist.cols != cols)
		needs_update = 1;

	ret = slist_resize(&blist->slist, rows, cols, screen_cols);
	if (ret != 0)
		return (-1);

	blist->label_len = SLIST_LAST_COL(&blist->slist) + 1;

	if (needs_update)
		blist_changed_width(blist);

	return (0);
}

inline void blist_destroy(struct blist *blist) {
	if (blist == NULL)
		return;

	slist_destroy(&blist->slist);
	memset(blist, 0, sizeof(*blist));
}

inline void blist_draw(struct blist *blist) {
	slist_draw(&blist->slist);
	blist_draw_cursor(blist, 1);
}

inline int blist_refresh(struct blist *blist) {
	return (slist_refresh(&blist->slist));
}

void blist_draw_border(struct blist *blist, int border_state) {
	attr_t border_attr;

	if (border_state == 0)
		border_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_NOFOCUS));
    else
		border_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_FOCUS));

	slist_clear_bot(&blist->slist);

	wattrset(blist->slist.win, border_attr);
	box(blist->slist.win, ACS_VLINE, ACS_HLINE);
	wattrset(blist->slist.win, 0);
}

inline void blist_draw_cursor(struct blist *blist, int status) {
	attr_t curs_attr = 0;

	if (status)
		curs_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_SELECTOR));

	slist_draw_cursor(&blist->slist, curs_attr);
}

inline void blist_draw_line(struct blist *blist, struct slist_cell *cell) {
	slist_draw_line(&blist->slist, cell);
}

// Called when you change a format string for the buddy list

void blist_changed_format(struct blist *blist) {
	dlist_t *cur;

	cur = blist->slist.cells;
	while (cur != NULL) {
		blist_update_label(blist, cur);
		cur = slist_node_next(cur);
	}
}

void blist_changed_width(struct blist *blist) {
	dlist_t *cur;

	cur = blist->slist.cells;
	while (cur != NULL) {
		struct slist_cell *cell = cur->data;

		free(cell->label);
		cell->label = xmalloc(sizeof(chtype) * blist->label_len);

		blist_update_label(blist, cur);
		cur = slist_node_next(cur);
	}
}
