/*
** pork_command.c - interface to commands typed by the user
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_status.h>
#include <pork_alias.h>
#include <pork_opt.h>
#include <pork_conf.h>
#include <pork_perl.h>
#include <pork_timer.h>
#include <pork_command.h>
#include <pork_command_defs.h>

extern struct screen screen;

/*
** Note that the "struct command" arrays are arranged in alphabetical
** order.  They have to be like that.
*/

/*
** The / command set.
*/

static const struct command command[] = {
	{ "",			cmd_send			},
	{ "alias",		cmd_alias			},
	{ "away",		cmd_away			},
	{ "bind",		cmd_bind			},
	{ "blist",		cmd_blist			},
	{ "buddy",		cmd_buddy			},
	{ "connect",	cmd_connect			},
	{ "disconnect", cmd_disconnect		},
	{ "echo",		cmd_echo			},
	{ "email",		cmd_email			},
	{ "help",		cmd_help			},
	{ "history",	cmd_history			},
	{ "idle",		cmd_idle			},
	{ "input",		cmd_input			},
	{ "invite",		cmd_invite			},
	{ "join",		cmd_join			},
	{ "msg",		cmd_msg				},
	{ "part",		cmd_part			},
	{ "passwd",		cmd_passwd			},
	{ "perl",		cmd_perl			},
	{ "perl_load",	cmd_perl_load		},
	{ "profile",	cmd_profile,		},
	{ "query",		cmd_query			},
	{ "quit",		cmd_quit			},
	{ "refresh",	cmd_refresh			},
	{ "save",		cmd_save			},
	{ "scroll",		cmd_scroll			},
	{ "set",		cmd_set				},
	{ "timer",		cmd_timer			},
	{ "unalias",	cmd_unalias			},
	{ "unbind",		cmd_unbind			},
	{ "win",		cmd_window			},
};

/*
** /input commands
*/

static const struct command input_command[] = {
	{ "backspace",				cmd_input_bkspace			},
	{ "clear",					cmd_input_clear				},
	{ "clear_next_word",		cmd_input_clear_next		},
	{ "clear_prev_word",		cmd_input_clear_prev		},
	{ "clear_to_end",			cmd_input_clear_to_end		},
	{ "clear_to_start",			cmd_input_clear_to_start	},
	{ "delete",					cmd_input_delete			},
	{ "end",					cmd_input_end				},
	{ "focus_next",				cmd_input_focus_next		},
	{ "insert",					cmd_input_insert			},
	{ "left",					cmd_input_left				},
	{ "next_word",				cmd_input_next_word			},
	{ "prev_word",				cmd_input_prev_word			},
	{ "prompt",					cmd_input_prompt			},
	{ "right",					cmd_input_right				},
	{ "send",					cmd_input_send				},
	{ "start",					cmd_input_start				},
};

USER_COMMAND(cmd_input_bkspace) {
	struct imwindow *imwindow = cur_window();

	input_bkspace(imwindow->input);
}

USER_COMMAND(cmd_input_clear) {
	struct imwindow *imwindow = cur_window();

	input_clear_line(imwindow->input);
}

USER_COMMAND(cmd_input_clear_prev) {
	struct imwindow *imwindow = cur_window();

	input_clear_prev_word(imwindow->input);
}

USER_COMMAND(cmd_input_clear_next) {
	struct imwindow *imwindow = cur_window();

	input_clear_next_word(imwindow->input);
}

USER_COMMAND(cmd_input_clear_to_end) {
	struct imwindow *imwindow = cur_window();

	input_clear_to_end(imwindow->input);
}

USER_COMMAND(cmd_input_focus_next) {
	struct imwindow *imwindow = cur_window();

	imwindow_switch_focus(imwindow);
	imwindow_blist_draw(imwindow);
}

USER_COMMAND(cmd_input_clear_to_start) {
	struct imwindow *imwindow = cur_window();

	input_clear_to_start(imwindow->input);
}

USER_COMMAND(cmd_input_delete) {
	struct imwindow *imwindow = cur_window();

	input_delete(imwindow->input);
}

USER_COMMAND(cmd_input_end) {
	struct imwindow *imwindow = cur_window();

	input_end(imwindow->input);
}

USER_COMMAND(cmd_input_insert) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL)
		input_insert_str(imwindow->input, args);
}

USER_COMMAND(cmd_input_left) {
	struct imwindow *imwindow = cur_window();

	input_move_left(imwindow->input);
}

USER_COMMAND(cmd_input_prev_word) {
	struct imwindow *imwindow = cur_window();

	input_prev_word(imwindow->input);
}

USER_COMMAND(cmd_input_prompt) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL) {
		input_set_prompt(imwindow->input, NULL);
		screen_status_msg("Input prompt cleared");
	} else {
		if (input_set_prompt(imwindow->input, args) == -1)
			screen_err_msg("Your prompt is too long for the screen");
		else {
			screen_status_msg("Input prompt set to %s", args);
		}
	}
}

USER_COMMAND(cmd_input_next_word) {
	struct imwindow *imwindow = cur_window();

	input_next_word(imwindow->input);
}

USER_COMMAND(cmd_input_right) {
	struct imwindow *imwindow = cur_window();

	input_move_right(imwindow->input);
}

USER_COMMAND(cmd_input_send) {
	struct imwindow *imwindow = cur_window();
	struct input *input = imwindow->input;
	static int recursion;

	/*
	** This is kind of a hack, but it's necessary if the client
	** isn't to crash when someone types "/input send"
	*/

	if (recursion == 1 && args == NULL)
		return;

	recursion = 1;

	if (args != NULL) {
		xstrncpy(input->input_buf, args, sizeof(input->input_buf));
		input->len = strlen(input->input_buf);
	}

	if (input->len > 0) {
		char *input_str = xstrdup(input->input_buf);

		input_history_add(input);
		input_clear_line(input);

		if (*input_str == opt_get_char(OPT_CMDCHARS))
			run_command(&input_str[1], CMDSET_MAIN);
		else {
			if (!imwindow->priv->connected)
				screen_err_msg("You're not logged in");
			else if (imwindow->type == TYPE_IM) {
				imwindow_send_im(imwindow, input_str);
				aim_send_msg(imwindow->priv, imwindow->target, input_str);
			}
		}

		free(input_str);
	}

	recursion = 0;
}

USER_COMMAND(cmd_input_start) {
	struct imwindow *imwindow = cur_window();

	input_home(imwindow->input);
}

/*
** /win commands
*/

static const struct command scroll_command[] = {
	{ "by",						cmd_scroll_by				},
	{ "down",					cmd_scroll_down				},
	{ "end",					cmd_scroll_end				},
	{ "page_down",				cmd_scroll_pgdown			},
	{ "page_up",				cmd_scroll_pgup				},
	{ "start",					cmd_scroll_start			},
	{ "up",						cmd_scroll_up				},
};

USER_COMMAND(cmd_scroll_by) {
	int lines;
	char *end;
	struct imwindow *imwindow = cur_window();

	if (args == NULL)
		return;

	lines = strtol(args, &end, 0);
	if (*end != '\0') {
		screen_err_msg("Bad number of lines: %s", args);
		return;
	}

	imwindow_scroll_by(imwindow, lines);
}

USER_COMMAND(cmd_scroll_down) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_down(imwindow);
}

USER_COMMAND(cmd_scroll_end) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_end(imwindow);
}

USER_COMMAND(cmd_scroll_pgdown) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_page_down(imwindow);
}

USER_COMMAND(cmd_scroll_pgup) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_page_up(imwindow);
}

USER_COMMAND(cmd_scroll_start) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_start(imwindow);
}

USER_COMMAND(cmd_scroll_up) {
	struct imwindow *imwindow = cur_window();

	imwindow_scroll_up(imwindow);
}

/*
** window commands.
*/

static const struct command window_command[] = {
	{ "bind",					cmd_window_bind		},
	{ "clear",					cmd_window_clear	},
	{ "close",					cmd_window_close	},
	{ "erase",					cmd_window_erase	},
	{ "list",					cmd_window_list		},
	{ "next",					cmd_window_next		},
	{ "prev",					cmd_window_prev		},
	{ "rename",					cmd_window_rename	},
	{ "renumber",				cmd_window_renumber	},
	{ "set",					cmd_window_set		},
	{ "swap",					cmd_window_swap		},
};

USER_COMMAND(cmd_window_bind) {
	struct imwindow *imwindow = cur_window();
	int ret;

	if (args == NULL || blank_str(args)) {
		if (imwindow->priv != NULL && imwindow->priv->username != NULL) {
			screen_status_msg("This window is bound to account: %s",
				imwindow->priv->username);
		} else
			screen_status_msg("This window is bound to no account");

		return;
	}

	ret = imwindow_bind_acct(imwindow, args);
	if (ret == -1)
		screen_err_msg("Account %s isn't signed on", args);
	else {
		screen_status_msg("This window is now bound to account: %s",
			imwindow->priv->username);
	}
}

USER_COMMAND(cmd_window_clear) {
	imwindow_clear(cur_window());
}

USER_COMMAND(cmd_window_close) {
	screen_close_window(cur_window());
}

USER_COMMAND(cmd_window_erase) {
	imwindow_erase(cur_window());
}

USER_COMMAND(cmd_window_list) {
	char buf[4096];
	size_t i = 0;
	size_t len = sizeof(buf);
	int ret;
	dlist_t *cur;
	static const char *win_types[] = { "Status", "IM", "Chat Room" };

	screen_status_msg("Window List:");
	cur = screen.window_list;
	do {
		struct imwindow *imwindow = cur->data;

		ret = snprintf(&buf[i], len, "Window refnum %u: Name=%s, Type=%s, Target=%s\n",
				imwindow->refnum, imwindow->name,
				win_types[imwindow->type], imwindow->target);
		i += ret;
		len -= ret;

		cur = cur->next;
	} while (cur != screen.window_list);

	screen_print_multiline(cur_window(), buf, 0);
}

USER_COMMAND(cmd_window_next) {
	screen_cycle_fwd();
}


USER_COMMAND(cmd_window_prev) {
	screen_cycle_bak();
}

USER_COMMAND(cmd_window_rename) {
	struct imwindow *imwindow = cur_window();

	imwindow_rename(imwindow, xstrdup(args));
}

USER_COMMAND(cmd_window_renumber) {
	u_int32_t num;
	char *end;
	struct imwindow *imwindow = cur_window();

	if (args == NULL) {
		screen_status_msg("This is window %u", imwindow->refnum);
		return;
	}

	num = strtoul(args, &end, 0);
	if (*end != '\0') {
		screen_err_msg("Bad window number: %s", args);
	}

	screen_renumber(imwindow, num);
}

USER_COMMAND(cmd_window_set) {
	char *variable;
	char *value;
	int opt;
	struct imwindow *imwindow = cur_window();

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		wopt_print(imwindow);
		return;
	}

	opt = wopt_find(variable);
	allcaps(variable);
	if (opt == -1) {
		screen_err_msg("Unknown variable: %s", variable);
		return;
	}

	value = strsep(&args, "\0");
	if (value == NULL || blank_str(value)) {
		wopt_print_var(imwindow, opt, "is set to");
		return;
	}

	if (wopt_set(imwindow, opt, value) == -1) {
		screen_err_msg("Bad argument for %s: %s", variable, value);
	} else {
		wopt_print_var(imwindow, opt, "set to");
	}
}

/*
** History Manipulation.
*/

static const struct command history_command[] = {
	{ "clear",					cmd_history_clear			},
	{ "next",					cmd_history_next			},
	{ "prev",					cmd_history_prev			},
};

USER_COMMAND(cmd_history_clear) {
	struct imwindow *imwindow = cur_window();

	input_history_clear(imwindow->input);
}

USER_COMMAND(cmd_history_next) {
	struct imwindow *imwindow = cur_window();

	input_history_next(imwindow->input);
}

USER_COMMAND(cmd_history_prev) {
	struct imwindow *imwindow = cur_window();

	input_history_prev(imwindow->input);
}

/*
** /buddy commands
*/

static const struct command buddy_command[] = {
	{ "add",			cmd_buddy_add			},
	{ "add_block",		cmd_buddy_add_block		},
	{ "add_group",		cmd_buddy_add_group		},
	{ "add_permit",		cmd_buddy_add_permit	},
	{ "alias",			cmd_buddy_alias			},
	{ "away",			cmd_buddy_away			},
	{ "clear_block",	cmd_buddy_clear_block	},
	{ "clear_permit",	cmd_buddy_clear_permit	},
	{ "permdeny_mode",	cmd_buddy_permdeny_mode	},
	{ "print_block",	cmd_buddy_print_block	},
	{ "print_list",		cmd_buddy_print_list	},
	{ "print_permit",	cmd_buddy_print_permit	},
	{ "profile",		cmd_buddy_profile		},
	{ "remove",			cmd_buddy_remove		},
	{ "remove_block",	cmd_buddy_remove_block	},
	{ "remove_group",	cmd_buddy_remove_group	},
	{ "remove_permit",	cmd_buddy_remove_permit	},
	{ "report_idle",	cmd_buddy_report_idle	},
	{ "search",			cmd_buddy_search		},
	{ "warn",			cmd_buddy_warn			},
	{ "warn_anon",		cmd_buddy_warn_anon		},
};

USER_COMMAND(cmd_buddy_add) {
	char *screen_name;
	char *group_name;
	struct bgroup *group;
	struct buddy *buddy;
	struct imwindow *imwindow = cur_window();

	screen_name = strsep(&args, " ");
	if (screen_name == NULL) {
		screen_err_msg("syntax is /buddy add <user> <group>");
		return;
	}

	buddy = buddy_find(imwindow->priv, screen_name);
	if (buddy != NULL) {
		screen_err_msg("%s is already a member of the group %s",
			screen_name, buddy->group->name);
		return;
	}

	group_name = strsep(&args, " ");
	if (group_name == NULL) {
		screen_err_msg("syntax is /buddy add <user> <group>");
		return;
	}

	group = group_find(imwindow->priv, group_name);
	if (group == NULL) {
		screen_err_msg("The group %s does not exist", group_name);
		return;
	}

	buddy_add(imwindow->priv, screen_name, group, 1);
}

USER_COMMAND(cmd_buddy_add_block) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_block(imwindow->priv, args, 1) == -1)
		screen_err_msg("%s is already blocked", args);
	else
		screen_status_msg("%s added to block list", args);
}

USER_COMMAND(cmd_buddy_add_permit) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (buddy_add_permit(imwindow->priv, args, 1) == -1)
		screen_err_msg("%s is already permitted", args);
	else
		screen_status_msg("%s added to permit list", args);
}

USER_COMMAND(cmd_buddy_add_group) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	group_add(imwindow->priv, args, 1);
}

USER_COMMAND(cmd_buddy_alias) {
	struct buddy *buddy;
	char *buddy_name;
	char *alias;
	struct imwindow *imwindow = cur_window();

	buddy_name = strsep(&args, " ");
	if (buddy_name == NULL)
		return;

	alias = strsep(&args, " ");
	if (alias == NULL || blank_str(alias))
		return;

	buddy = buddy_find(imwindow->priv, buddy_name);
	if (buddy == NULL) {
		screen_err_msg("%s is not on your buddy list", buddy_name);
		return;
	}

	screen_status_msg("alias: %s -> %s", buddy_name,alias);
	buddy_alias(imwindow->priv, buddy, alias);
}

USER_COMMAND(cmd_buddy_away) {
	char nname[NUSER_LEN];
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args)) {
		aim_get_away_msg(imwindow->priv, imwindow->priv->username);
		return;
	}

	normalize(args, nname, sizeof(nname));
	aim_get_away_msg(imwindow->priv, nname);
}

USER_COMMAND(cmd_buddy_clear_block) {
	struct imwindow *imwindow = cur_window();
	dlist_t *cur = imwindow->priv->buddy_pref->block_list;

	while (cur != NULL) {
		dlist_t *next = cur->next;

		buddy_remove_block(imwindow->priv, cur->data, 1);
		cur = next;
	}
}

USER_COMMAND(cmd_buddy_clear_permit) {
	struct imwindow *imwindow = cur_window();
	dlist_t *cur = imwindow->priv->buddy_pref->permit_list;

	while (cur != NULL) {
		dlist_t *next = cur->next;

		buddy_remove_permit(imwindow->priv, cur->data, 1);
		cur = next;
	}
}

USER_COMMAND(cmd_buddy_permdeny_mode) {
	struct imwindow *imwindow = cur_window();
	int mode = -1;

	if (args != NULL) {
		char *end;
		int new_mode;

		new_mode = strtol(args, &end, 10);
		if (*end == '\0')
			mode = new_mode;
	}

	mode = aim_permdeny_mode(imwindow->priv, mode);
	screen_status_msg("Your permit/deny mode is %d", mode);
}

USER_COMMAND(cmd_buddy_print_block) {
	struct imwindow *imwindow = cur_window();

	buddy_print_block(imwindow->priv);
}

USER_COMMAND(cmd_buddy_print_list) {
	struct imwindow *imwindow = cur_window();

	buddy_print_list(imwindow->priv);
}

USER_COMMAND(cmd_buddy_profile) {
	char nname[NUSER_LEN];
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args)) {
		aim_whois(imwindow->priv, imwindow->priv->username);
		return;
	}

	normalize(args, nname, sizeof(nname));
	aim_whois(imwindow->priv, nname);
}

USER_COMMAND(cmd_buddy_print_permit) {
	struct imwindow *imwindow = cur_window();

	buddy_print_permit(imwindow->priv);
}

USER_COMMAND(cmd_buddy_remove_permit) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_permit(imwindow->priv, args, 1) != 0)
		screen_err_msg("%s isn't on your permit list", args);
	else
		screen_status_msg("%s removed from permit list", args);
}

USER_COMMAND(cmd_buddy_remove_block) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove_block(imwindow->priv, args, 1) != 0)
		screen_err_msg("%s isn't on your block list", args);
	else
		screen_status_msg("%s removed from block list", args);
}

USER_COMMAND(cmd_buddy_remove) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (buddy_remove(imwindow->priv, args, 1) != 0)
		screen_err_msg("%s isn't on your buddy list", args);
	else
		screen_status_msg("%s removed from buddy list", args);
}

USER_COMMAND(cmd_buddy_remove_group) {
	struct imwindow *imwindow = cur_window();

	if (args == NULL || blank_str(args))
		return;

	if (group_remove(imwindow->priv, args, 1) != 0)
		screen_err_msg("No such group: %s", args);
	else
		screen_status_msg("Removed group: %s", args);
}

USER_COMMAND(cmd_buddy_report_idle) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL && imwindow->priv->connected) {
		char *end;
		u_int32_t mode;

		mode = strtoul(args, &end, 10);
		if (*end != '\0') {
			screen_err_msg("Invalid number: %s", args);
			return;
		}

		aim_report_idle(imwindow->priv, mode);
	}

	if (imwindow->priv->buddy_pref == NULL)
		return;

	screen_status_msg("The reporting of idle time is %s",
		(imwindow->priv->buddy_pref->report_idle ? "enabled" : "disabled"));
}

USER_COMMAND(cmd_buddy_search) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->connected && args != NULL)
		aim_usersearch_address(&priv->aim_session, priv->bos_conn, args);
}

USER_COMMAND(cmd_buddy_warn) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args))
		return;

	aim_warn(imwindow->priv, args);
}

USER_COMMAND(cmd_buddy_warn_anon) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL || blank_str(args))
		return;

	aim_warn_anon(imwindow->priv, args);
}

/*
** /blist commands
*/

static const struct command blist_command[] = {
	{ "add_block",		cmd_blist_add_block		},
	{ "add_permit",		cmd_blist_add_permit	},
	{ "away",			cmd_blist_away			},
	{ "collapse",		cmd_blist_collapse		},
	{ "down",			cmd_blist_down			},
	{ "end",			cmd_blist_end			},
	{ "goto",			cmd_blist_goto			},
	{ "hide",			cmd_blist_hide			},
	{ "page_down",		cmd_blist_pgdown		},
	{ "page_up",		cmd_blist_pgup			},
	{ "profile",		cmd_blist_profile		},
	{ "refresh",		cmd_blist_refresh		},
	{ "remove",			cmd_blist_remove		},
	{ "remove_block",	cmd_blist_remove_block	},
	{ "remove_permit",	cmd_blist_remove_permit	},
	{ "select",			cmd_blist_select		},
	{ "show",			cmd_blist_show			},
	{ "start",			cmd_blist_start			},
	{ "up",				cmd_blist_up			},
	{ "warn",			cmd_blist_warn			},
	{ "warn_anon",		cmd_blist_warn_anon		},
	{ "width",			cmd_blist_width			},
};

USER_COMMAND(cmd_blist_add_block) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_add_block(buddy->nname);
}

USER_COMMAND(cmd_blist_add_permit) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_add_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_away) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_away(buddy->nname);
}

USER_COMMAND(cmd_blist_collapse) {
	struct slist_cell *cell;
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (args == NULL)
		cell = blist_get_cursor(priv->blist);
	else {
		dlist_t *node;
		struct bgroup *gr = group_find(imwindow->priv, args);

		if (gr == NULL)
			return;

		node = gr->blist_line;
		if (node == NULL || node->data == NULL)
			return;
		cell = node->data;
	}

	if (cell == NULL || cell->type != TYPE_LIST_CELL)
		return;

	blist_collapse_group(priv->blist, cell->data);
}

USER_COMMAND(cmd_blist_down) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	blist_cursor_down(priv->blist);
}

USER_COMMAND(cmd_blist_end) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	blist_cursor_end(priv->blist);
}

USER_COMMAND(cmd_blist_goto) {
	if (args != NULL)
		cmd_query(args);
	else {
		struct slist_cell *cell;
		struct imwindow *imwindow = cur_window();
		struct buddy *buddy;

		cell = blist_get_cursor(imwindow->priv->blist);
		if (cell == NULL)
			return;

		if (cell->type != TYPE_LIST_CELL)
			return;

		buddy = cell->data;
		cmd_query(buddy->nname);
	}
}

USER_COMMAND(cmd_blist_hide) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	imwindow_blist_hide(imwindow);
}

USER_COMMAND(cmd_blist_pgdown) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	blist_cursor_pgdown(priv->blist);
}

USER_COMMAND(cmd_blist_pgup) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	blist_cursor_pgup(priv->blist);
}

USER_COMMAND(cmd_blist_profile) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_profile(buddy->nname);
}

USER_COMMAND(cmd_blist_refresh) {
	struct imwindow *imwindow = cur_window();

	imwindow_blist_draw(imwindow);
}

USER_COMMAND(cmd_blist_remove) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_remove(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_block) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_remove_block(buddy->nname);
}

USER_COMMAND(cmd_blist_remove_permit) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_remove_permit(buddy->nname);
}

USER_COMMAND(cmd_blist_select) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell;

	cell = blist_get_cursor(priv->blist);
	if (cell == NULL)
		return;

	if (cell->type == TYPE_LIST_CELL)
		cmd_blist_collapse(NULL);
	else {
		struct buddy *buddy = cell->data;

		cmd_blist_goto(buddy->nname);
	}
}

USER_COMMAND(cmd_blist_show) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	imwindow_blist_show(imwindow);
}

USER_COMMAND(cmd_blist_start) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	if (priv->blist == NULL) {
		screen_status_msg("Log in first");
		return;
	}

	blist_cursor_start(priv->blist);
}

USER_COMMAND(cmd_blist_up) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;

	blist_cursor_up(priv->blist);
}

USER_COMMAND(cmd_blist_warn) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_warn(buddy->nname);
}

USER_COMMAND(cmd_blist_warn_anon) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv = imwindow->priv;
	struct slist_cell *cell = blist_get_cursor(priv->blist);
	struct buddy *buddy;

	if (cell == NULL || cell->type == TYPE_LIST_CELL)
		return;
	buddy = cell->data;

	cmd_buddy_warn_anon(buddy->nname);
}

USER_COMMAND(cmd_blist_width) {
	u_int32_t new_len;
	char *end;
	struct imwindow *imwindow;

	if (args == NULL)
		return;

	new_len = strtoul(args, &end, 10);
	if (*end != '\0') {
		screen_err_msg("Error: invalid width: %s", args);
		return;
	}

	imwindow = cur_window();
	screen_blist_width(imwindow->priv->blist, new_len);
}

static const struct command timer_command[] = {
	{ "add",			cmd_timer_add			},
	{ "del",			cmd_timer_del			},
	{ "del_refnum",		cmd_timer_del_refnum	},
	{ "list",			cmd_timer_list			},
	{ "purge",			cmd_timer_purge			},
};

USER_COMMAND(cmd_timer_add) {
	char *p;
	char *end;
	u_int32_t interval;
	int persist = 0;

	if (args == NULL)
		return;

	p = strsep(&args, " ");
	if (p == NULL)
		return;

	interval = strtoul(p, &end, 10);
	if (*end != '\0') {
		screen_err_msg("Invalid timer interval: %s", p);
		return;
	}

	p = strsep(&args, " ");
	if (p == NULL)
		return;

	if (*p == '1')
		persist = 1;

	p = strsep(&args, "\0");
	if (p == NULL)
		return;

	timer_add(&screen.timer_list, p, interval, persist);
}

USER_COMMAND(cmd_timer_del) {
	int ret;

	if (args == NULL)
		return;

	ret = timer_del(&screen.timer_list, args);
	if (ret == -1)
		screen_err_msg("No timer for \"%s\" was found", args);
	else
		screen_status_msg("Timer for \"%s\" was removed", args);	
}

USER_COMMAND(cmd_timer_del_refnum) {
	u_int32_t refnum;
	char *end;
	int ret;

	if (args == NULL)
		return;

	refnum = strtoul(args, &end, 10);
	if (*end != '\0') {
		screen_err_msg("Bad timer refnum: %s", args);
		return;
	}

	ret = timer_del_refnum(&screen.timer_list, refnum);
	if (ret == -1)
		screen_err_msg("No timer with refnum %u was found", refnum);
	else
		screen_status_msg("Timer with refnum %u was removed", refnum);	
}

USER_COMMAND(cmd_timer_list) {
	screen_status_msg("INTERVAL\t\tPERSISTENT\t\tREFNUM\t\tCOMMAND");
	dlist_iterate(screen.timer_list, print_timer, NULL);
}

USER_COMMAND(cmd_timer_purge) {
	if (screen.timer_list != NULL) {
		timer_destroy(&screen.timer_list);
		screen_status_msg("All timers have been removed");
	}
}

static const struct {
	const struct command *set;
	size_t elem;
	char *type;
} command_set[] = {
	{ command,			array_elem(command),			"" 			},
	{ window_command,	array_elem(window_command),		"window "	},
	{ history_command,	array_elem(history_command),	"history "	},
	{ input_command,	array_elem(input_command),		"input "	},
	{ scroll_command,	array_elem(scroll_command),		"scroll "	},
	{ buddy_command,	array_elem(buddy_command),		"buddy "	},
	{ blist_command,	array_elem(blist_command),		"blist "	},
	{ timer_command,	array_elem(timer_command),		"timer "	},
};

/*
** Main command set.
*/

USER_COMMAND(cmd_alias) {
	char *alias;
	char *str;

	alias = strsep(&args, " ");
	if (alias == NULL || blank_str(alias)) {
		hash_iterate(&screen.alias_hash, print_alias, NULL);
		return;
	}

	str = strsep(&args, "\0");
	if (str == NULL || blank_str(str)) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_status_msg("%s is aliased to %s",
				lalias->alias, lalias->str);
		} else
			screen_status_msg("There is no alias for %s", alias);

		return;
	}

	if (alias_add(&screen.alias_hash, alias, str) == 0) {
		struct alias *lalias = alias_find(&screen.alias_hash, alias);

		if (lalias != NULL) {
			screen_status_msg("%s is aliased to %s",
				lalias->alias, lalias->str);
			return;
		}
	}

	screen_status_msg("Error adding alias for %s", alias);
}

USER_COMMAND(cmd_away) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	aim_set_away(imwindow->priv, args);
}

USER_COMMAND(cmd_bind) {
	int key;
	char *func;
	char *key_str;
	struct imwindow *imwindow = cur_window();
	struct key_binds *target_binds = imwindow->active_binds;
	struct binding *binding;

	key_str = strsep(&args, " ");
	if (key_str == NULL || blank_str(key_str)) {
		hash_iterate(&target_binds->hash, print_binding, NULL);
		return;
	}

	if (key_str[0] == '-' && key_str[1] != '\0') {
		if (!strcasecmp(key_str, "-b") || !strcasecmp(key_str, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(key_str, "-m") || !strcasecmp(key_str, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_status_msg("Bad bind flag: %s", key_str);
			return;
		}

		key_str = strsep(&args, " ");

		if (key_str == NULL || blank_str(key_str)) {
			hash_iterate(&target_binds->hash, print_binding, NULL);
			return;
		}
	}

	key = bind_get_keycode(key_str);
	if (key == -1) {
		screen_status_msg("Bad keycode: %s", key_str);
		return;
	}

	func = strsep(&args, "\0");
	if (func != NULL) {
		if (*func == opt_get_char(OPT_CMDCHARS) && *(func + 1) != '\0')
			func++;
		if (blank_str(func))
			func = NULL;
	}

	if (func == NULL) {
		binding = bind_find(target_binds, key);
		if (binding != NULL)
			screen_status_msg("%s is bound to %s", key_str, binding->binding);
		else
			screen_status_msg("%s is not bound", key_str);

		return;
	}

	bind_add(target_binds, key, func);
	binding = bind_find(target_binds, key);
	if (binding != NULL) {
		screen_status_msg("%s is bound to %s", key_str, binding->binding);
		return;
	}

	screen_status_msg("Error binding %s", key_str);
}

USER_COMMAND(cmd_connect) {
	struct imwindow *imwindow = cur_window();
	struct aim_priv *priv;

	if (args != NULL && !blank_str(args)) {
		struct aim_acct *acct;

		acct = pork_acct_add(args);
		if (acct == NULL) {
			screen_status_msg("You're already connected as %s", args);
			return;
		}

		priv = acct->priv;
		if (imwindow->priv == &screen.null_priv) {
			cmd_window_bind(acct->username);

			if (!imwindow->blist_visible) {
				if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BUDDY_LIST))
					wopt_set(imwindow, WOPT_SHOW_BUDDY_LIST, "1");
			}

		}
	} else {
		if (imwindow->priv->connected) {
			cmd_window_bind(imwindow->priv->username);
			return;
		}

		priv = imwindow->priv;
	}

	if (priv->passwd == NULL) {
		char buf[128];

		screen_get_passwd(buf, sizeof(buf));
		if (buf[0] == '\0') {
			/* EH */
			screen_status_msg("There was an error reading your password");
			return;
		}

		priv->passwd = xstrdup(buf);
		memset(buf, 0, sizeof(buf));
	}

	if (aim_login(priv) == -1)
		screen_status_msg("Unable to login as %s", priv->username);
}

USER_COMMAND(cmd_echo) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL)
		screen_win_msg(imwindow, 0, 0, 1, "%s", args);
}

USER_COMMAND(cmd_disconnect) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL && !blank_str(args)) {
		if (pork_acct_del(args) != 0)
			screen_status_msg("You're not logged in as %s", args);
	} else {
		if (pork_acct_del(imwindow->priv->username) != 0) {
			screen_status_msg("Account %s is not logged in",
				imwindow->priv->username);
		}
	}
}

USER_COMMAND(cmd_help) {
}

USER_COMMAND(cmd_idle) {
	char *end;
	u_int32_t idle_secs;
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (args == NULL) {
		aim_set_idle(imwindow->priv, 0);
		return;
	}

	idle_secs = strtoul(args, &end, 10);
	if (*end != '\0') {
		screen_status_msg("Bad argument: %s", args);
		return;
	}

	aim_set_idle(imwindow->priv, idle_secs);
}

USER_COMMAND(cmd_invite) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_join) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_msg) {
	struct imwindow *conv_window;
	char *target;
	char *msg;
	struct imwindow *imwindow = cur_window();

	if (args == NULL)
		return;

	if (!imwindow->priv->connected) {
		screen_status_msg("You're not logged in");
		return;
	}

	target = strsep(&args, " ");
	if (target == NULL)
		return;
	msg = strsep(&args, "\0");
	if (msg == NULL)
		return;

	conv_window = imwindow_find(target);
	if (conv_window == NULL || conv_window->type != TYPE_IM) {
		conv_window = screen_new_window(imwindow->priv, target);
		if (conv_window == NULL)
			return;
		screen_goto_window(conv_window->refnum);
	}

	aim_send_msg(conv_window->priv, target, msg);
	imwindow_send_im(conv_window, msg);
}

USER_COMMAND(cmd_part) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_query) {
	struct imwindow *imwindow = cur_window();

	if (args != NULL && !blank_str(args)) {
		struct imwindow *conv_window;

		conv_window = imwindow_find(args);
		if (conv_window == NULL)
			conv_window = screen_new_window(imwindow->priv, args);

		screen_goto_window(conv_window->refnum);
	} else
		screen_close_window(imwindow);
}

USER_COMMAND(cmd_quit) {
	pork_exit(0, NULL);
}

USER_COMMAND(cmd_refresh) {
	screen_refresh();
}

USER_COMMAND(cmd_save) {
	if (save_global_config() == 0)
		screen_status_msg("Your configuration has been saved");
	else
		screen_err_msg("There was an error saving your configuration");
}

USER_COMMAND(cmd_send) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}

	if (imwindow->type == TYPE_IM && args != NULL)
		aim_send_msg(imwindow->priv, imwindow->target, args);
}

USER_COMMAND(cmd_unbind) {
	char *binding;
	struct imwindow *imwindow = cur_window();
	struct key_binds *target_binds = imwindow->active_binds;
	int c;

	binding = strsep(&args, " ");
	if (binding == NULL || blank_str(binding))
		return;

	if (binding[0] == '-' && binding[1] != '\0') {
		if (!strcasecmp(binding, "-b") || !strcasecmp(binding, "-buddy"))
			target_binds = &screen.binds.blist;
		else if (!strcasecmp(binding, "-m") || !strcasecmp(binding, "-main"))
			target_binds = &screen.binds.main;
		else {
			screen_status_msg("Bad unbind flag: %s", binding);
			return;
		}

		binding = strsep(&args, " ");

		if (binding == NULL || blank_str(binding))
			return;
	}

	c = bind_get_keycode(binding);
	if (c == -1) {
		screen_status_msg("Bad keycode: %s", binding);
		return;
	}

	if (bind_remove(target_binds, c) == -1)
		screen_status_msg("There is no binding for %s", binding);
	else
		screen_status_msg("Binding for %s removed", binding);
}

USER_COMMAND(cmd_unalias) {
	char *alias = strsep(&args, " ");

	if (alias == NULL)
		return;

	if (alias_remove(&screen.alias_hash, alias) == -1)
		screen_status_msg("%s: No such alias", alias);
	else
		screen_status_msg("Alias %s removed", alias);
}

USER_COMMAND(cmd_email) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_passwd) {
	struct imwindow *imwindow = cur_window();

	if (!imwindow->priv->connected) {
		screen_err_msg("You're not connected");
		return;
	}
}

USER_COMMAND(cmd_perl) {
	size_t num_args = 0;
	char *p = args;
	char **perl_args;
	char *function;
	char *orig;
	size_t i = 0;

	if (args == NULL)
		return;

	p = args;
	while (*p == ' ')
		p++;

	if (*p == '$')
		p++;
	args = p;
	orig = p;

	function = args;
	p = strchr(function, '(');
	if (p != NULL) {
		*p++ = '\0';
		args = p;
	}

	p = strchr(function, ' ');
	if (p != NULL) {
		*p = '\0';
		args = p + 1;
	}

	if (args == orig) {
		execute_perl(function, NULL);
		return;
	}

	p = args;
	while ((p = strchr(p, ',')) != NULL) {
		++num_args;
		++p;
	}
	num_args += 2;

	p = strchr(args, ')');
	if (p != NULL)
		*p = '\0';

	perl_args = xcalloc(num_args, sizeof(char *));

	while ((p = strsep(&args, ",")) != NULL) {
		char *s;

		while (*p == ' ' || *p == '\t')
			p++;

		s = strrchr(p, ' ');
		if (s != NULL && blank_str(s)) {
			while (*s == ' ')
				s--;
			*(s + 1) = '\0';
		}

		perl_args[i++] = p;
	}

	perl_args[i] = NULL;

	execute_perl(function, perl_args);
	free(perl_args);
}

USER_COMMAND(cmd_perl_load) {
	int ret;

	if (args == NULL)
		return;

	ret = perl_load_file(args);
	if (ret != 0)
		screen_err_msg("Error: The file %s couldn't be loaded", args);
}

USER_COMMAND(cmd_profile) {
	struct imwindow *imwindow = cur_window();

	aim_set_profile(imwindow->priv, args);
}

USER_COMMAND(cmd_window_swap) {
	u_int32_t num;
	char *end;

	if (args == NULL)
		return;

	num = strtoul(args, &end, 0);
	if (*end != '\0')
		return;

	screen_goto_window(num);
}

USER_COMMAND(cmd_window) {
	if (args != NULL)
		run_command(args, CMDSET_WIN);
}

USER_COMMAND(cmd_buddy) {
	if (args != NULL)
		run_command(args, CMDSET_BUDDY);
}

USER_COMMAND(cmd_blist) {
	if (args != NULL)
		run_command(args, CMDSET_BLIST);
}

USER_COMMAND(cmd_input) {
	if (args != NULL)
		run_command(args, CMDSET_INPUT);
}

USER_COMMAND(cmd_history) {
	if (args != NULL)
		run_command(args, CMDSET_HISTORY);
}

USER_COMMAND(cmd_scroll) {
	if (args != NULL)
		run_command(args, CMDSET_SCROLL);
}

USER_COMMAND(cmd_timer) {
	if (args != NULL)
		run_command(args, CMDSET_TIMER);
}

USER_COMMAND(cmd_set) {
	char *variable;
	char *value;
	int opt;

	variable = strsep(&args, " ");
	if (variable == NULL || blank_str(variable)) {
		opt_print();
		return;
	}

	allcaps(variable);
	opt = opt_find(variable);
	if (opt == -1) {
		screen_err_msg("Unknown variable: %s", variable);
		return;
	}

	value = strsep(&args, "\0");
	if (value == NULL || blank_str(value)) {
		opt_print_var(opt, "is set to");
		return;
	}

	if (opt_set(opt, value) == -1) {
		screen_err_msg("Bad argument for %s: %s", variable, value);
	} else {
		opt_print_var(opt, "set to");
	}
}

int run_command(char *str, u_int32_t set) {
	static char buf[INPUT_BUFFER_LEN * 4];
	struct alias *alias;
	char *cmd_str;
	struct command *cmd;

	xstrncpy(buf, str, sizeof(buf));
	str = buf;

	cmd_str = strsep(&str, " ");
	if (set == CMDSET_MAIN) {
		alias = alias_find(&screen.alias_hash, cmd_str);

		if (alias != NULL) {
			char *p;
			char *alias_cmd;
			char *alias_args;
			static char alias_buf[INPUT_BUFFER_LEN];

			xstrncpy(alias_buf, alias->str, sizeof(alias_buf));

			p = alias_buf;
			alias_cmd = strsep(&p, " ");
			alias_args = strsep(&p, "\0");

			if (alias_args != NULL) {
				if (str == NULL) {
					str = alias_args;
					cmd_str = alias_cmd;
				} else {
					size_t aarg_len = strlen(alias_args);
					size_t str_len = strlen(str);

					if (aarg_len + str_len + 1 < sizeof(buf) - str_len) {
						memmove(&str[aarg_len + 1], str, str_len + 1);
						str[aarg_len] = ' ';
						memcpy(str, alias_args, aarg_len);
						cmd_str = alias_cmd;
					} else
						screen_err_msg("The alias is too long");
				}
			} else
				cmd_str = alias_cmd;
		}
	}

	cmd = bsearch(cmd_str, command_set[set].set,
			command_set[set].elem, sizeof(struct command), cmd_compare);

	if (cmd == NULL) {
		screen_err_msg("Unknown %scommand: %s",
			command_set[set].type, cmd_str);

		return (-1);
	}

	cmd->cmd(str);
	return (0);
}

static void print_binding(void *data, void *nothing __notused) {
	struct binding *binding = data;
	char key_name[32];

	bind_get_keyname(binding->key, key_name, sizeof(key_name));
	screen_status_msg("%s is bound to %s", key_name, binding->binding);
}

static void print_alias(void *data, void *nothing __notused) {
	struct alias *alias = data;

	screen_status_msg("%s is aliased to %s", alias->alias, alias->str);
}

static int cmd_compare(const void *l, const void *r) {
	char *key = (char *) l;
	struct command *cmd = (struct command *) r;

	return (strcasecmp(key, cmd->name));
}

static void print_timer(void *data, void *nothing __notused) {
	struct timer_entry *timer = data;

	screen_status_msg("%d\t\t\t%s\t\t\t%u\t\t%s", timer->interval,
		(timer->persist ? "yes" : "no"), timer->refnum, timer->command);
}
