/*
** pork_imwindow.c - interface for manipulating conversation/chat/info windows.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_color.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_format.h>

extern struct screen screen;

struct imwindow *imwindow_new(	u_int32_t rows,
								u_int32_t cols,
								u_int32_t refnum,
								u_int32_t type,
								struct aim_priv *priv,
								char *target)
{
	struct imwindow *imwindow;
	WINDOW *swin;
	struct input *input;
	char nname[NUSER_LEN];

	swin = newwin(rows, cols, 0, 0);
	if (swin == NULL)
		return (NULL);

	normalize(target, nname, sizeof(nname));

	imwindow = xcalloc(1, sizeof(*imwindow));
	imwindow->refnum = refnum;
	imwindow->type = type;
	imwindow->target = xstrdup(nname);
	imwindow->name = xstrdup(target);
	imwindow->active_binds = &screen.binds.main;
	imwindow->priv = priv;
	imwindow->priv->ref_count++;

	wopt_init(imwindow, target);

	swindow_init(&imwindow->swindow, swin, rows,
		cols, imwindow->opts);

	/*
	** Use a separate input handler.
	*/

	if (wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT)) {
		input = xmalloc(sizeof(*input));
		input_init(input, cols);
	} else
		input = &screen.input;
	imwindow->input = input;

	if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BUDDY_LIST)) {
		if (priv->blist != NULL)
			imwindow_blist_show(imwindow);
	}

	return (imwindow);
}

inline void imwindow_rename(struct imwindow *imwindow, char *new_name) {
	free(imwindow->name);
	imwindow->name = new_name;
}

inline void imwindow_resize(struct imwindow *imwindow,
							u_int32_t rows,
							u_int32_t cols)
{
	swindow_resize(&imwindow->swindow, rows, cols);
	swindow_scroll_to_end(&imwindow->swindow);
}

inline int imwindow_blist_refresh(struct imwindow *imwindow) {
	struct aim_priv *priv = imwindow->priv;

	if (!imwindow->blist_visible)
		return (-1);

	if (!priv->blist->slist.dirty)
		return (-1);

	blist_draw_border(priv->blist, imwindow->input_focus);
	blist_refresh(priv->blist);
	return (0);
}

inline int imwindow_refresh(struct imwindow *imwindow) {
	int was_dirty_win = -1;
	int was_dirty_blist = -1;

	if (imwindow->swindow.visible && imwindow->swindow.dirty) {
		was_dirty_win = 0;
		wnoutrefresh(imwindow->swindow.win);
		imwindow->swindow.dirty = 0;
	}

	was_dirty_blist = imwindow_blist_refresh(imwindow);
	if (!was_dirty_win || !was_dirty_blist)
		return (0);

	return (-1);
}

void imwindow_blist_show(struct imwindow *imwindow) {
	u_int32_t new_width;

	if (imwindow->blist_visible)
		return;

	imwindow->blist_visible = 1;
	imwindow->priv->blist->slist.dirty = 1;
	new_width = imwindow->swindow.cols - imwindow->priv->blist->slist.cols;
	imwindow_resize(imwindow, imwindow->swindow.rows, new_width);
}

void imwindow_blist_hide(struct imwindow *imwindow) {
	u_int32_t new_width;

	if (!imwindow->blist_visible)
		return;

	imwindow->blist_visible = 0;
	imwindow->priv->blist->slist.dirty = 1;
	new_width = imwindow->swindow.cols + imwindow->priv->blist->slist.cols;
	imwindow_resize(imwindow, imwindow->swindow.rows, new_width);
}

void imwindow_blist_draw(struct imwindow *imwindow) {
	struct blist *blist = imwindow->priv->blist;

	if (blist == NULL || !imwindow->blist_visible)
		return;

	blist_draw(imwindow->priv->blist);
}

void imwindow_destroy(struct imwindow *imwindow) {
	swindow_destroy(&imwindow->swindow);

	if (wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT)) {
		input_destroy(imwindow->input);
		free(imwindow->input);
	}

	imwindow->priv->ref_count--;
	if (imwindow->priv->ref_count == 0) {
		/* XXX: might want to delete it here (config option). */
	}

	wopt_destroy(imwindow);
	free(imwindow->name);
	free(imwindow->target);
	free(imwindow);
}

void imwindow_switch_focus(struct imwindow *imwindow) {
	if (!imwindow->blist_visible)
		return;

	if (imwindow->input_focus == BINDS_MAIN) {
		imwindow->input_focus = BINDS_BUDDY;
		imwindow->active_binds = &screen.binds.blist;
	} else {
		imwindow->input_focus = BINDS_MAIN;
		imwindow->active_binds = &screen.binds.main;
	}
}

struct imwindow *imwindow_find(const char *target) {
	dlist_t *list_start = screen.window_list;
	dlist_t *cur = list_start;
	char nname[NUSER_LEN];

	normalize(target, nname, sizeof(nname));

	do {
		struct imwindow *imwindow = cur->data;

		if (!strcasecmp(imwindow->target, nname))
			return (imwindow);

		cur = cur->next;
	} while (cur != list_start);

	return (NULL);
}

struct imwindow *imwindow_find_name(const char *name) {
	dlist_t *list_start = screen.window_list;
	dlist_t *cur = list_start;

	do {
		struct imwindow *imwindow = cur->data;

		if (!strcasecmp(imwindow->target, name))
			return (imwindow);

		cur = cur->next;
	} while (cur != list_start);

	return (NULL);
}

struct imwindow *imwindow_find_refnum(u_int32_t refnum) {
	dlist_t *list_start = screen.window_list;
	dlist_t *cur = list_start;

	do {
		struct imwindow *imwindow = cur->data;

		if (imwindow->refnum == refnum)
			return (imwindow);

		cur = cur->next;
	} while (cur != list_start);

	return (NULL);
}

/*
** Print the instant message that we've just sent
** to the window "win"
*/

void imwindow_send_im(struct imwindow *win, char *str) {
	char *im_send_str = opt_get_str(OPT_FORMAT_IM_SEND);
	char *buf;
	chtype *ch;
	size_t len;
	struct aim_priv *priv = win->priv;

	len = strlen(im_send_str) + strlen(str) + strlen(priv->username) + 1;
	buf = xmalloc(len);
	make_im(im_send_str, buf, len, priv->username, str);
	ch = xmalloc(sizeof(chtype) * len);
	plaintext_to_cstr(buf, ch, len);
	free(buf);

	swindow_input(&win->swindow, ch);
	free(ch);
}

/*
** Print the instant message "msg" from "source" to
** the window specified by "win"
*/

void imwindow_recv_im(	struct imwindow *win,
						const char *source,
						const char *msg)
{
	char *p;
	size_t len;
	char *buf;
	chtype *ch;
	char *im_recv_str = opt_get_str(OPT_FORMAT_IM_RECV);

	p = strchr(msg, '\n');
	if (p != NULL)
		*p++ = '\0';

	len = strlen(msg) + strlen(im_recv_str) + strlen(source) + 1;
	buf = xmalloc(len);
	make_im(im_recv_str, buf, len, source, msg);
	ch = xmalloc(sizeof(chtype) * len);

	plaintext_to_cstr(buf, ch, len);
	free(buf);
	swindow_add(&win->swindow, ch, 1);
	free(ch);

	if (p != NULL)
		screen_print_multiline(win, p, 0);

	if (wopt_get_bool(win->opts, WOPT_BEEP_ON_OUTPUT))
		beep();
}

/*
** Bind the screen name "user" to the current window.
*/

int imwindow_bind_acct(struct imwindow *imwindow, char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];
	struct aim_priv *old_priv = imwindow->priv;

	normalize(user, nname, sizeof(nname));

	node = pork_acct_find(nname);
	if (node == NULL)
		return (-1);

	acct = node->data;

	if (old_priv != acct->priv) {
		imwindow->priv->ref_count--;
		imwindow->priv = acct->priv;
		imwindow->priv->ref_count++;

		if (imwindow->blist_visible) {
			imwindow_blist_draw(imwindow);
			imwindow_blist_refresh(imwindow);
		}
	}

	return (0);
}

inline void imwindow_scroll_up(struct imwindow *imwindow) {
	swindow_scroll_up(&imwindow->swindow);
}

inline void imwindow_scroll_down(struct imwindow *imwindow) {
	swindow_scroll_down(&imwindow->swindow);
}

inline void imwindow_scroll_by(struct imwindow *imwindow, int lines) {
	swindow_scroll_by(&imwindow->swindow, lines);
}

inline void imwindow_scroll_page_up(struct imwindow *imwindow) {
	swindow_scroll_by(&imwindow->swindow, -imwindow->swindow.rows);
}

inline void imwindow_scroll_page_down(struct imwindow *imwindow) {
	swindow_scroll_by(&imwindow->swindow, imwindow->swindow.rows);
}

inline void imwindow_scroll_start(struct imwindow *imwindow) {
	swindow_scroll_to_start(&imwindow->swindow);
}

inline void imwindow_scroll_end(struct imwindow *imwindow) {
	swindow_scroll_to_end(&imwindow->swindow);
}

inline void imwindow_clear(struct imwindow *imwindow) {
	swindow_clear(&imwindow->swindow);
}

inline void imwindow_erase(struct imwindow *imwindow) {
	swindow_erase(&imwindow->swindow);
}
